#!/usr/local/bin/perl
#
# NAME:
#	md5s.pl - update MD5.list if present.
#
# SYNOPSIS:
#	md5s.pl [options] ["dir" ...]
#	md5s.pl [options] -M "dir"/"MD5list"
#
# DESCRIPTION:
#	Install_files(8) will use cached md5 signatures if available
#	to minimize NFS traffic.  The md5 signatures are stored in
#	files named MD5.list in various ``root'' directories thoughout
#	the configs tree.
#	
#	This tool is intended to be run regularly on the configs host.
#	Its role is to help ensure the MD5.lists are up to date.
#	It will find all the MD5.lists below "dir" and update them for
#	any file bigger than "bytes" that is newer than the
#	specified date (default is the mtime of MD5.list).
#	
#	The file size threshold is chosen to minimize NFS traffic when
#	computing md5 signatures while at the same time ensuring upto
#	date signatures for files which may change frequently.  Thus
#	the md5 signature is not cached for file which can be read in
#	a single read.
#
#	If "dir" is omitted it starts at the top of the "CONFIGS"
#	tree.
#
#	Options:
#
#	-M "MD5list"
#		We will update files called "MD5list" (default MD5.list).
#		If "MD5list" contains any '/' we assume only a single
#		list is to be updated.
#
#	-m "md5"
#		Allows the location of "md5" to be specified.
#
#	-b "bytes"
#		We only "md5" files bigger than "bytes" (default 8k).
#		
#	-s "since"
#		Specifies the number of seconds since the unix epoch.
#		Any file bigger than "bytes" with an mtime > "since"
#		will be fed to "md5".  If "since" names a file, we use
#		its mtime, otherwise if "since" begins with a digit
#		less than 5, it is assumed to be a
#		YYYYmmdd[HH[MM[SS]]] date spec.  Using "since" of 0
#		will force all files bigger than "bytes" to be fed to
#		"md5". 
#		
#	-f	Normally when an "MD5list" is found in a sub-directory
#		of "dir", a separate process is forked to handle the
#		tree from that point on.  This option causes
#		sub-directories conaining an MD5.list to be skipped.
#		
#	-v	Be verbose.  Tell folk what is happening.
#
#	If the "MD5list" is empty, all files > "bytes" will be fed to
#	"md5".
#
# AUTHOR:
#	Simon J. Gerraty <sjg@crufty.net>

# RCSid:
#	$Id: md5s.pl,v 1.11 2004/04/17 05:33:04 sjg Exp $
#
#	@(#) Copyright (c) 1994 Simon J. Gerraty
#
#	This file is provided in the hope that it will
#	be of use.  There is absolutely NO WARRANTY.
#	Permission to copy, redistribute or otherwise
#	use this file is hereby granted provided that 
#	the above copyright notice and this notice are
#	left intact. 
#      
#	Please send copies of changes and bug-fixes to:
#	sjg@crufty.net
#

if ($0 =~ m,^(.+)/([^/]+)$,) {
  $Mydir = $1;
  $Myname = $2;
} else {
  $Mydir = '.';
  $Myname = $0;
}
chop($Mydir = `cd $Mydir; pwd`) if ($Mydir !~ m,^/,);
$Myname =~ s,\.pl$,,;

if ($Mydir =~ m,^(/.+)/([^/]+)$,) {
  $CONFIGS = $1;
}

require 'getopts.pl';

push(@INC, ($Mydir, "$CONFIGS/usr/local/lib/perl"));

require "$CONFIGS/share/bin/only1.pl";
&Only1;
require 'basename.pl';
require "find.pl";

$opt_M = 'MD5.list';
$opt_m = 'md5';
$opt_b = 8192;
$DIR = 'none';			# ensure test in wanted fails.

&Getopts("M:m:b:s:vfnV:");

if ($opt_V ne '') {
  $opt_v = '-v';
}
if ($opt_s ne '') {
  if (-f $opt_s) {
    @s = stat($opt_s);
    $opt_s = $s[9];		# file's mtime
  } elsif ($opt_s =~ m/^[1-4]/) {
    # unlikely a valid time, assume a YYYYmmdd[HH[MM[SS]]] spec
    require 'mktime.pl';
    $opt_s = &mktime($opt_s);
  }
}

&main;
exit 0;

sub main {
  if ($opt_M =~ m,^(.*)/([^/]+)$,) {
    @ARGV = ($1);
    if (! -f $opt_M && $opt_f ne '') {
      open(M, "> $opt_M") || die;
      close(M);
    }
    $opt_M = $2;
    $opt_f = '-f';		# don't go anywhere else
  }
  if (scalar(@ARGV) == 0) {
    push(@ARGV, $CONFIGS);
  }
  print "$Myname\[$$\]: find(@ARGV)\n" if ($opt_v ne '');
  find(@ARGV);
  close M if ($DIR ne 'none');
  for (;;) {
    last if (wait() < 0);
  }
  print "$Myname\[$$\]: done.\n" if ($opt_v ne '');
}

sub wanted {
  local($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,
   $atime,$mtime,$ctime,$blksize,$blocks)
    = stat($_);
  $base = $_;
  if (-d _) {
    if ($base =~ m,^(CVS|tmp|junk|obj\..*)$,) {
      $prune = 1;
    } elsif (-f "$base/$opt_M") {
      ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,
       $atime,$mtime,$ctime,$blksize,$blocks) = stat("$base/$opt_M");
      if ($base eq '.') {
	close M if ($DIR ne 'none');
	$DIR = $dir;
      } else {
	&flush(STDOUT);
	&flush(STDERR);
	&flush(M) if ($DIR ne 'none');
	# do NOT find() here if $base is a symlink
	if ($opt_f eq '' && ! -l $base) {
	  if ($opt_n ne '') {
	    print "$Myname\[$$\]: fork(), find($dir/$base)\n";
	  } elsif (fork() == 0) {
	    close M if ($DIR ne 'none');
	    $DIR = 'none';
	    print "$Myname\[$$\]: fork(), find($dir/$base)\n" if ($opt_v ne '');
	    &find("$dir/$base");
	    close M if ($DIR ne 'none');
	    wait;
	    print "$Myname\[$$\]: done\n" if ($opt_v ne '');
	    exit 0;
	  }
	}
	$prune = 1;
	return;
      }
      if ($size == 0 || $opt_s eq '0') {
	$Cmd = ($size == 0) ? '' : "rename('$DIR/$opt_M', '$DIR/$opt_M.bak'); ";
	$Cmd .= "open(M, '| (cd $DIR && xargs $opt_m > $opt_M )')";
	$since = 0;
      } else {
	$Cmd = "open(M, '| (cd $DIR && xargs $CONFIGS/umd5)')";

	if ($opt_s ne '') {
	  $since = $opt_s;
	} else {
	  $since = $mtime;
	}
      }
    }
  } elsif ($size > $opt_b && $base ne 'core' &&
	   $base !~ m/([~#]|\.(list|old|bak|core))$/) {
    if ("$dir/$base" =~ m,^$DIR/(.*), && $mtime > $since) {
      if ($Cmd ne '') {
	if ($opt_n ne '' || $opt_v ne '') {
	  print "$Myname\[$$\]: $Cmd, size > $opt_b && mtime > $since\n";
	}
	if ($opt_n eq '') {
	  eval $Cmd || die;
	}
	$Cmd = '';
      }
      print M "$1\n" if ($opt_n eq '');
      if ($opt_V eq '2') {
	print "$Myname\[$$\]: $1\n";
      } elsif ($opt_V ne '') {
	print "$1\n";
      }
    }
  }
}

sub flush {
  local(*H) = @_;

  select(H);
  $| = 1;
  print H "";
  $| = 0;
  select(stdout);
}
